
import unittest

from testnum import TestNumericalTaylor
from testsym import TestSymbolicTaylor
from testutils import TestTaylorUtils
from testdiffcalc import TestDiffCalc


def main():
    unittest.main()


if __name__ == "__main__":
    main()