from setuptools import setup

setup(
    name='ohtaylor',
    version='1.0.0',
    packages=['ohtaylor'],
    url='https://gitlab.com/mochan/ohtaylor',
    license='Apache 2.0',
    author='Mochan Shrestha',
    author_email='',
    description='Taylor Series calculator'
)
